import cadquery as cq

# all measurements in mm
#
# x = width
# y = length = the longest axis of the phone
# z = height = the shortest axis of the phone
#
# -z represents the screen of the phone
# +z represents the rear of the phone

# phone body
body_length = 161
body_width = 76.6
body_height = 9.2
# about the radii:
# - if too small, the phone will fit the case but the case will be "boxier" and have a bit of wiggle room
# - if too large, then slotting the phone into the case will be difficult.
# TPU is stretchy so not a huge deal, but best to err *slightly* on the side of under-sizing the radii
body_rad_xy = 8.6
body_rad_front_z = 1.8
body_rad_back_z = body_rad_front_z  #< TODO: measure this

# offset from top of phone to top of volume button
volume_min_y = 27.0
volume_min_z = 3.5
volume_length = 19.8
volume_height = 2.0
volume_width = 0.5

# offset from top of phone to the top of power button
power_min_y = 54.0
power_min_z = 3.5
power_length = 10.0
power_height = 2.0
power_width = 0.5

# offset from left edge of phone to left edge of 3.5mm audio jack
aux_min_x = 23.0
aux_min_z = 3.8

usb_width = 9.0
usb_height = 3.5
usb_min_z = 5.1

camera_min_y = 4.0
camera_min_x = 41.6
camera_width = 22.0
camera_length = 10.0
camera_depth = 2.0  #< guess

def body():
    return (
        cq.Workplane("front")
        .box(body_width, body_length, body_height, centered=False)
        .tag("body_box")
        .faces(">>X").tag("body_right")
        .faces(tag="body_box")
        .faces("<<Y").tag("body_top")
        .faces(tag="body_box")
        .faces(">>Y").tag("body_bottom")
        .faces(tag="body_box")
        .faces(">>Z").tag("body_back")
        .faces(tag="body_box")
        .edges("|Z")
        .fillet(body_rad_xy)
        .edges("(|X or |Y)")
        .fillet(body_rad_front_z)
        .tag("body")
    )

def volume(body):
    return (
        body
        .faces(tag="body_right")
        .workplane()
        .move(volume_min_y, volume_min_z)
        .box(volume_length, volume_height, volume_width, centered=False, combine=False)
        .tag("volume_box")
        .edges("|X")
        .fillet(volume_height/3)  #< XXX: cadquery doesn't allow a full 1/2 fillet
        .edges("(not |X) and (not <X)")
        .fillet(volume_width/2)
        .tag("volume")
    )

def power(body):
    return (
        body
        .faces(tag="body_right")
        .workplane()
        .move(power_min_y, power_min_z)
        .box(power_length, power_height, power_width, centered=False, combine=False)
        .tag("power_box")
        .edges("|X")
        .fillet(power_height/3)  #< XXX: cadquery doesn't allow a full 1/2 fillet
        .edges("(not |X) and (not <X)")
        .fillet(power_width/2)
        .tag("power")
    )

def camera(body):
    return (
        body
        .faces(tag="body_back")
        .workplane()
        .move(camera_min_x, camera_min_y)
        .box(camera_width, camera_length, camera_depth, centered=False, combine=False)
        .tag("camera_box")
        .edges("|Z")
        .fillet(camera_length/3)
        .edges("(not |Z) and (not < Z)")
        .fillet(camera_depth/2)
        .tag("camera")
    )

def aux(body):
    return (
        body
        .faces(tag="body_top")
        .workplane()
        .move(aux_min_x, aux_min_z)
        .cylinder(radius=3.5/2, height=10, direct=(0, 0, -1), centered=False, combine=False)
        .tag("aux")
    )

def usb(body):
    return (
        body
        .faces(tag="body_bottom")
        .workplane(offset=-10)
        .move(-body_width/2, usb_min_z)
        .box(usb_width, usb_height, 10, centered=(True, False, False), combine=False)
        .tag("usb_box")
        .edges("|Y")
        .fillet(usb_height/3)
        .tag("usb")
    )

def PinePhone():
    body_ = body()
    phone = body_
    phone = phone.union(volume(body_))
    phone = phone.union(power(body_))
    phone = phone.union(camera(body_))
    phone = phone.cut(aux(body_))
    phone = phone.cut(usb(body_))
    return phone
